@@ -1,6 +1,6 @@
-// Copyright (C) 2003 Samy Bengio (bengio@idiap.ch)
+// Copyright (C) 2003--2004 Samy Bengio (bengio@idiap.ch)
 //                
-// This file is part of Torch 3.
+// This file is part of Torch 3.1.
 //
 // All rights reserved.
 // 
@@ -35,14 +35,19 @@
 
 namespace Torch {
 
-HMM::HMM(int n_states_, Distribution **states_, real** transitions_) : Distribution(states_[1]->n_inputs,n_states_*n_states_)
+HMM::HMM(int n_states_, Distribution **states_, real** transitions_, int n_shared_states_,Distribution **shared_states_) : Distribution(states_[1]->n_inputs,n_states_*n_states_)
 {
   n_states = n_states_;
   states = states_;
   transitions = transitions_;
+
+  // if given, these represent the "real" states, while "states" are
+  // only pointers to shared_states. note that shared_states should all
+  // exist (not like states which is null for states[0] and states[n_states-1])
+  shared_states = shared_states_;
+  n_shared_states = n_shared_states_;
   
   addBOption("initialize", &initialize , true, "initialize the model before training");
-	addBOption("linear segmentation", &linear_segmentation, false, "linear segmentation to initialize the states");
 	addBOption("linear segmentation", &linear_segmentation, false, "linear segmentation to initialize the states");
 	addROption("prior transitions", &prior_transitions , 1e-3, "minimum weights for each gaussians");
   
@@ -266,17 +271,36 @@ void HMM::logViterbi(Sequence* inputs)
     }
   }
   // now recall the state sequence
-  viterbi_sequence->frames[inputs->n_frames-1][0] = last_arg_viterbi;
-  for (int f=inputs->n_frames-2;f>=0;f--) {
-    viterbi_sequence->frames[f][0] = (real)(arg_viterbi->frames[f+1][(int)(viterbi_sequence->frames[f+1][0])]);
+  if (log_probability > LOG_ZERO) {
+    viterbi_sequence->frames[inputs->n_frames-1][0] = last_arg_viterbi;
+    for (int f=inputs->n_frames-2;f>=0;f--) {
+      viterbi_sequence->frames[f][0] = (real)(arg_viterbi->frames[f+1][(int)(viterbi_sequence->frames[f+1][0])]);
+    }
+  } else {
+    warning("sequence impossible to train: probably too short for target");
+    for (int f=0;f<inputs->n_frames;f++)
+      viterbi_sequence->frames[f][0] = -1;
+    log_probability = 0;
   }
 }
 
 void HMM::logProbabilities(Sequence *inputs)
 {
-  for (int f=0;f<inputs->n_frames;f++) {
-    for (int i=1;i<n_states-1;i++)
-      log_probabilities_s->frames[f][i] = states[i]->frameLogProbability(f, inputs->frames[f]);
+  if (n_shared_states == 0) {
+    for (int f=0;f<inputs->n_frames;f++) {
+      for (int i=1;i<n_states-1;i++) {
+        log_probabilities_s->frames[f][i] = states[i]->frameLogProbability(f, inputs->frames[f]);
+      }
+    }
+  } else {
+    for (int f=0;f<inputs->n_frames;f++) {
+      for (int i=0;i<n_shared_states;i++) {
+        shared_states[i]->frameLogProbability(f, inputs->frames[f]);
+      }
+      for (int i=1;i<n_states-1;i++) {
+        log_probabilities_s->frames[f][i] = states[i]->log_probabilities->frames[f][0];
+      }
+    }
   }
 }
 
@@ -390,9 +414,13 @@ void HMM::viterbiAccPosteriors(Sequence *inputs, real log_posterior)
   real p = exp(log_posterior);
   for (int f=0;f<inputs->n_frames;f++) {
     int i = (int)(viterbi_sequence->frames[f][0]);
-    states[i]->frameEMAccPosteriors(f, inputs->frames[f],log_posterior);
-    int j = (int)arg_viterbi->frames[f][i];
-    transitions_acc[i][j] += p;
+    if (i>=0) {
+      states[i]->frameEMAccPosteriors(f, inputs->frames[f],log_posterior);
+      int j = (int)arg_viterbi->frames[f][i];
+      if (j>0) {
+        transitions_acc[i][j] += p;
+      }
+    }
   }
   // attention, il me manque le premier et le dernier arg_viterbi...
 }
