@@ -1,6 +1,6 @@
-// Copyright (C) 2003 Samy Bengio (bengio@idiap.ch)
+// Copyright (C) 2003--2004 Samy Bengio (bengio@idiap.ch)
 //                
-// This file is part of Torch 3.
+// This file is part of Torch 3.1.
 //
 // All rights reserved.
 // 
@@ -40,7 +40,7 @@ SpeechHMM::SpeechHMM(int n_models_, HMM **models_, LexiconInfo* lex_, EMTrainer*
   max_n_states = 1;
   initial_models_trainer_measurers = NULL;
   addROption("prior transitions", &prior_transitions , 1e-3, "minimum weights for each gaussians");
-  addBOption("targets are phonemes", &phoneme_targets, LOG_ONE, "for initialization, targets are often phonemes instead of words");
+  addBOption("targets are phonemes", &phoneme_targets, false, "for initialization, targets are often phonemes instead of words");
   model_trainer = model_trainer_;
   lexicon = lex_;
 
@@ -127,14 +127,29 @@ void SpeechHMM::saveXFile(XFile *file)
 void SpeechHMM::setTargets(Sequence* t)
 {
   if (!phoneme_targets && add_to_targets > 0) {
-    targets->resize(t->n_frames+add_to_targets);
+    if (targets->frame_size != t->frame_size) {
+      allocator->free(targets);
+      targets = new(allocator) Sequence(t->n_frames+add_to_targets,t->frame_size);
+    } else
+      targets->resize(t->n_frames+add_to_targets);
     int j=0;
-    if (lexicon->vocabulary->sent_start_index >= 0)
-      targets->frames[j++][0] = lexicon->vocabulary->sent_start_index;
-    for (int i=0;i<t->n_frames;i++)
-      targets->frames[j++][0] = t->frames[i][0];
-    if (lexicon->vocabulary->sent_end_index >= 0)
+    if (lexicon->vocabulary->sil_index >= 0) {
+      if (t->frame_size > 1)
+        targets->frames[j][1] = 1; // silence lasts 1 frame
+      targets->frames[j++][0] = lexicon->vocabulary->sil_index;
+    }
+    for (int i=0;i<t->n_frames;i++,j++) {
+      for (int k=0;k<t->frame_size;k++) {
+        targets->frames[j][k] = t->frames[i][k];
+      }
+    }
+    if (lexicon->vocabulary->sent_end_index >= 0) {
+      if (t->frame_size > 1) {
+        targets->frames[j][1] = t->frames[t->n_frames-1][1];
+        targets->frames[j-1][1] = t->frames[t->n_frames-1][1]-1;
+      }
       targets->frames[j++][0] = lexicon->vocabulary->sent_end_index;
+    }
   } else
     targets = t;
 }
@@ -162,8 +177,10 @@ void SpeechHMM::splitDataSet(DataSet* data_, ExampleFrameSelectorDataSet** split
           // the alignment is given in terms of phonemes
           int bu = (int)begin_unit;
           int eu = (int)end_unit;
-          if (eu >= n_frames)
-            eu = n_frames - 1;
+          if (eu > n_frames) {
+            warning("The target alignment has been truncated from %d to %d",eu,n_frames-1);
+            eu = n_frames;
+          }
           int length = eu - bu;
           split_data_[unit]->addExample(i, bu, length, j,  1);
         } else {
@@ -187,8 +204,10 @@ void SpeechHMM::splitDataSet(DataSet* data_, ExampleFrameSelectorDataSet** split
               (models[phoneme]->n_states-2);
             int bu = (int)begin_phoneme;
             int eu = (int)end_phoneme;
-            if (eu >= n_frames)
+            if (eu >= n_frames) {
+              warning("SpeechHMM::splitData: The target alignment has been truncated from %d to %d",eu,n_frames-1);
               eu = n_frames - 1;
+            }
             int length = eu - bu;
             split_data_[phoneme]->addExample(i, bu, length, j,  1);
             begin_phoneme = end_phoneme;
@@ -214,8 +233,9 @@ void SpeechHMM::splitDataSet(DataSet* data_, ExampleFrameSelectorDataSet** split
               (models[phoneme]->n_states-2);
           int ib = (int)begin_phoneme;
           int ie = (int)end_phoneme;
-          if (ie >= n_frames)
+          if (ie >= n_frames) {
             ie = n_frames - 1;
+          }
           int length = ie - ib;
           if (length < 1) {
             warning("a segmentation produced a sequence of frames (%d) less than then number of states (%d) of the model",n_frames,tot_n_states);
@@ -250,8 +270,9 @@ void SpeechHMM::splitDataSet(DataSet* data_, ExampleFrameSelectorDataSet** split
               (models[phoneme]->n_states-2);
             int ib = (int)begin_phoneme;
             int ie = (int)end_phoneme;
-            if (ie >= n_frames)
+            if (ie >= n_frames) {
               ie = n_frames - 1;
+            }
             int length = ie - ib;
             if (length < 1) {
               warning("a segmentation produced a sequence of frames (%d) less than then number of states (%d) of the model",n_frames,tot_n_states);
@@ -274,7 +295,7 @@ void SpeechHMM::setDataSet(DataSet* data_)
 
   add_to_targets = 0;
   if (!phoneme_targets) {
-    if (lexicon->vocabulary->sent_start_index >= 0)
+    if (lexicon->vocabulary->sil_index >= 0)
       add_to_targets++;
     if (lexicon->vocabulary->sent_end_index >= 0)
       add_to_targets++;
@@ -311,7 +332,6 @@ void SpeechHMM::setDataSet(DataSet* data_)
 
 void SpeechHMM::addConnectionsBetweenWordsToModel(int word,int next_word, int current_state,int next_current_state, real log_n_next)
 {
-  int n_states_word = nStatesInWord(word);
   int n_states_word_pronun = 0;
   int n_pronuns = lexicon->vocab_to_lex_map[word].n_pronuns;
   for (int l=0;l<n_pronuns;l++) {
@@ -321,6 +341,8 @@ void SpeechHMM::addConnectionsBetweenWordsToModel(int word,int next_word, int cu
     int n_states_model = models[current_model]->n_states;
     int n_states_next_word_pronun = 0;
     int n_pronuns_next = lexicon->vocab_to_lex_map[next_word].n_pronuns;
+    int n_states_current_word_pronun = nStatesInWordPronunciation(word,l);
+    n_states_word_pronun += n_states_current_word_pronun;
     for (int m=0;m<n_pronuns_next;m++) {
       int next_lex_ind = lexicon->vocab_to_lex_map[next_word].pronuns[m];
       LexiconInfoEntry *next_lex_ent = &lexicon->entries[next_lex_ind];
@@ -329,17 +351,17 @@ void SpeechHMM::addConnectionsBetweenWordsToModel(int word,int next_word, int cu
       for (int j=1;j<n_states_model;j++) {
         if (models[current_model]->log_transitions[n_states_model-1][j] != LOG_ZERO) {
           for (int k=1;k<n_states_next_model-1;k++) {
-            if (models[next_model]->log_transitions[k][0] != LOG_ZERO)
-              log_transitions[next_current_state+n_states_next_word_pronun+k-1][current_state+n_states_word_pronun+n_states_word-n_states_model+1+j] = 
+            if (models[next_model]->log_transitions[k][0] != LOG_ZERO) {
+              log_transitions[next_current_state+n_states_next_word_pronun+k-1][current_state+n_states_word_pronun-n_states_model+1+j] = 
                 models[current_model]->log_transitions[n_states_model-1][j] +
                 models[next_model]->log_transitions[k][0] - log_n_next;
-              word_transitions[next_current_state+n_states_next_word_pronun+k-1][current_state+n_states_word_pronun+n_states_word-n_states_model+1+j] = true;
+              word_transitions[next_current_state+n_states_next_word_pronun+k-1][current_state+n_states_word_pronun-n_states_model+1+j] = true;
+            }
           }
         }
       }
       n_states_next_word_pronun += nStatesInWordPronunciation(next_word,m);
     }
-    n_states_word_pronun += nStatesInWordPronunciation(word,l);
   }
 }
 
@@ -355,6 +377,7 @@ int SpeechHMM::addWordToModel(int word, int current_state)
       // for each emitting state of the current model
       int n_states_model = models[current_model]->n_states;
       for (int j=1;j<n_states_model-1;j++,current_state++) {
+//printf("In word %d Loading model %d state %d into state %d\n",word, current_model, j, current_state);
         states[current_state] = models[current_model]->states[j];
         states_to_model_states[current_state] = j;
         states_to_model[current_state] = current_model;
@@ -412,7 +435,7 @@ void SpeechHMM::prepareTrainModel(Sequence* inputs)
   for (int i=0;i<n_states;i++) {
     for (int j=0;j<n_states;j++) {
       log_transitions[i][j] = LOG_ZERO;
-      word_transitions[i][j] = LOG_ZERO;
+      word_transitions[i][j] = false;
     }
   }
 
@@ -610,19 +633,23 @@ void SpeechHMM::viterbiAccPosteriors(Sequence *inputs, real log_posterior)
   real p = exp(log_posterior);
   for (int f=0;f<inputs->n_frames;f++) {
     int i = (int)viterbi_sequence->frames[f][0];
-    int model_to = states_to_model[i];
-    int state_to = states_to_model_states[i];
-    states[i]->frameEMAccPosteriors(f, inputs->frames[f], log_posterior);
-    int j = (int)arg_viterbi->frames[f][i];
-    // find the real transition
-    int model_from = states_to_model[j];
-    int state_from = states_to_model_states[j];
-    if (model_from == model_to) {
-      models[model_from]->transitions_acc[state_to][state_from] += p;
-    } else {
-      int last_state_from = models[model_from]->n_states-1;
-      models[model_from]->transitions_acc[last_state_from][state_from] += p;
-      models[model_to]->transitions_acc[state_to][0] += p;
+    if (i>=0) {
+      int model_to = states_to_model[i];
+      int state_to = states_to_model_states[i];
+      states[i]->frameEMAccPosteriors(f, inputs->frames[f], log_posterior);
+      int j = (int)arg_viterbi->frames[f][i];
+      // find the real transition
+      if (j>=0) {
+        int model_from = states_to_model[j];
+        int state_from = states_to_model_states[j];
+        if (model_from == model_to) {
+          models[model_from]->transitions_acc[state_to][state_from] += p;
+        } else {
+          int last_state_from = models[model_from]->n_states-1;
+          models[model_from]->transitions_acc[last_state_from][state_from] += p;
+          models[model_to]->transitions_acc[state_to][0] += p;
+        }
+      }
     }
   }
 }
@@ -698,18 +725,17 @@ void SpeechHMM::backward(Sequence *inputs, Sequence *alpha)
     }
   }
   // particular case of transitions from initial state
-  int f = 0;
-  int i = 0;
-  for (int j=1;j<n_states;j++) {
-    if (log_transitions[j][0] == LOG_ZERO ||
-      log_beta->frames[f][i] == LOG_ZERO)
+  for (int j=1;j<n_states-1;j++) {
+    if (log_transitions[j][0] == LOG_ZERO)
       continue;
     int model_to = states_to_model[j];
     int state_to = states_to_model_states[j];
-    real posterior_i_j_f = *alpha->frames[0] * exp(log_beta->frames[f][i] +
-      models[model_to]->log_transitions[state_to][0] - log_probability);
-    models[model_to]->log_transitions[state_to][0] += posterior_i_j_f;
-    for (int k=1;k<n_states;k++) {
+    real log_emit_j = models[model_to]->states[state_to]->log_probabilities->frames[0][0];
+    real posterior_i_j_f = *alpha->frames[0] * exp(log_beta->frames[0][j] +
+      log_emit_j + models[model_to]->log_transitions[state_to][0] - 
+      log_probability);
+    models[model_to]->dlog_transitions[state_to][0] += posterior_i_j_f;
+    for (int k=1;k<n_states-1;k++) {
       if (log_transitions[k][0] == LOG_ZERO)
         continue;
       int k_model_to = states_to_model[k];
@@ -719,9 +745,9 @@ void SpeechHMM::backward(Sequence *inputs, Sequence *alpha)
     }
   }
   // particular case of transitions to last state
-  f = inputs->n_frames;
-  i = n_states-1;
-  for (int j=1;j<n_states;j++) {
+  int f = inputs->n_frames-1;
+  int i = n_states-1;
+  for (int j=1;j<n_states-1;j++) {
     if (log_transitions[i][j] == LOG_ZERO ||
       log_alpha->frames[f][j] == LOG_ZERO)
       continue;
@@ -731,9 +757,9 @@ void SpeechHMM::backward(Sequence *inputs, Sequence *alpha)
     real posterior_i_j_f = *alpha->frames[0] * exp(log_alpha->frames[f][j] +
       models[model_from]->log_transitions[last_state_from][state_from] -
       log_probability);
-    models[model_from]->log_transitions[last_state_from][state_from] +=
+    models[model_from]->dlog_transitions[last_state_from][state_from] +=
       posterior_i_j_f;
-    for (int k=1;k<n_states;k++) {
+    for (int k=1;k<n_states-1;k++) {
       if (log_transitions[k][j] == LOG_ZERO)
         continue;
       int k_model_from = states_to_model[k];
